// This file is part of the QGLViewer library
// Copyright (C) 2002-2003  Gilles Debunne (Gilles.Debunne@imag.fr)

// http://www-imagis.imag.fr/Membres/Gilles.Debunne/CODE/QGLViewer


// libQGLViewer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// libQGLViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with libQGLViewer; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef QGLVIEWER_TRIANGLESETCONSTRAINT_H
#define QGLVIEWER_TRIANGLESETCONSTRAINT_H

#include "constraint.h"
#include "vec.h"

namespace qglviewer {
  /*! \brief A translation Constraint defined with respect to a set of triangles.
  
  Define your path (connected set of triangles) using addPoint() and addTriangle(). The frame will
  then be constrained to move on this path.
  
  Development version. Many things need to be cleared.
  
  See <a href="../examples/triSetConstraint.html">triSetConstraint</a> for an illustration. */
  class QGLVIEWER_EXPORT TriangleSetConstraint : public Constraint
  {
  public:
    /*! Create an empty TriangleSetConstraint. currentTriangle() set to 0. */
    TriangleSetConstraint() : currentTriangle_(0) , epsilon_(0.0) {};

    /*! Virtual empty destructor. */
    virtual ~TriangleSetConstraint() {};

    virtual void constrainTranslation(Vec&, Frame* const);
    // virtual void constrainRotation(Quaternion& , Frame* const);

    /*! Add a point to the triangle set. Use addTriangle() to create a triangle with this point.
      Returns the index of the point in the internal list.

      If this point is within a range of epsilon() around an already entered point, the point is
      not added. The returned index is the one of the first point. See indexOfPoint(). */
    unsigned int addPoint(const Vec& p);
    /*! Add a triangle between the three points whose indexes are given. Point are added with addPoint(), which
     returns their index. Returns the index of the created triangle (used by currentTriangle()).*/
    unsigned int addTriangle(const unsigned int i0, const unsigned int i1, const unsigned int i2);

    /*! Returns the index of point \p in the point list. Returns -1 if not found. Two points are
      declared identical if their distance is lower or equal than epsilon(). */
    int indexOfPoint(const Vec& p) const;

    /*! The index of the triangle the frame is currently on. Changes as the frame moves on the triangle set. */
    unsigned int currentTriangle() const { return currentTriangle_; };
    /*! Use this function to initialize the constraint. You should give the index of the triangle
      the frame is currently on (use Frame::setPosition() to move the frame if needed). If the constraint
      is lost and currentTriangle() is not valid, the frame will be positioned back to the center
      of the previous currentTriangle(). */
    void setCurrentTriangle(const unsigned int ct);

    /*! The current value of epsilon, used to determine identical points in addPoint(). */
    float epsilon() const { return epsilon_; };
    /*! Set the current value of epsilon, used to determine identical points in addPoint(). Default value is 0.0. */
    void setEpsilon(const float e) { epsilon_ = e; };
    
    /*! Returns the coordinates of the point at \p index. */
    Vec point(unsigned int index) const { return points_[index]; };
    /*! Returns the adjacent triangle of triangle \p index, along its edge \p edge. */
    int neighTriangle(unsigned int index, unsigned short edge) const { return triangles_[index].neighbor_[edge]; };
    /*! Returns the index of the \p nb point of triangle \p index. */
    unsigned int trianglePoint(unsigned int index, unsigned short nb) const { return triangles_[index].node_[nb]; };
    
  private:
    
    //! Internal private Triangle representation
    class Triangle
    {
    public:
      unsigned short indexOf(unsigned int i) const;
      unsigned int node_[3];
      int neighbor_[3];
    };
    
    unsigned int currentTriangle_;
    float epsilon_;    
    QValueVector<Triangle> triangles_;
    QValueVector<Vec> points_;

    QValueVector<unsigned int> trianglesContainingPoint(const unsigned int i) const;
  };

} // namespace qglviewer

#endif // QGLVIEWER_TRIANGLESETCONSTRAINT_H
